/*
 *  linux/drivers/i2c/i2c-armadillo9.c
 *
 *  Copyright (C) 2004-2005 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/miscdevice.h>
#include <linux/rtc.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "i2c-s3531a.h"
#include "i2c-at24cxx.h"

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "i2c Armadillo-9 driver"
#define LICENSE "GPL v2"

#define I2C_HW_B_ARMADILLO9 0x16

#define CHRDEV  "i2c-armadillo9"
#define VERSION CHRDEV ": " DESCRIPTION ", (C) 2004-2005 " AUTHOR

static struct i2c_client *rtc_client;

static inline int rtc_command(int cmd, void *data)
{
  int ret = -EIO;
  
  if (rtc_client) {
    ret = rtc_client->driver->command(rtc_client, cmd, data);
  }
  
  return ret;
}

static inline int rtc_sync()
{
  return rtc_command(RTC_SYNC, NULL);
}

static inline int get_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  int ret;
  
  memset(rtctm, 0, sizeof(struct rtc_time));
  
  ret = rtc_command(RTC_GETDATETIME, &rawtm);
  if (!ret) {
    rtctm->tm_year = rawtm.year + ((rawtm.year <= 69) ? 100 : 0);
    rtctm->tm_mon  = rawtm.mon - 1;
    rtctm->tm_mday = rawtm.mday;
    rtctm->tm_wday = rawtm.wday;
    rtctm->tm_hour = rawtm.hour;
    rtctm->tm_min  = rawtm.min;
    rtctm->tm_sec  = rawtm.sec;
  }
  
  return ret;
}

static inline int set_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  
  rawtm.year = rtctm->tm_year - ((rtctm->tm_year < 100) ? 0 : 100);
  rawtm.mon  = rtctm->tm_mon + 1;
  rawtm.mday = rtctm->tm_mday;
  rawtm.wday = rtctm->tm_wday;
  rawtm.hour = rtctm->tm_hour;
  rawtm.min  = rtctm->tm_min;
  rawtm.sec  = rtctm->tm_sec;
  
  return rtc_command(RTC_SETDATETIME, &rawtm);
}

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
  struct rtc_time rtctm;
  int ret;
  
  switch (cmd) {
  case RTC_ALM_READ:
    break;
    
  case RTC_ALM_SET:
    break;
    
  case RTC_RD_TIME:
    ret = get_rtc_time(&rtctm);
    if (ret) {
      return ret;
    }
    return copy_to_user((void *)arg, &rtctm, sizeof(rtctm)) ? -EFAULT : 0;
    
  case RTC_SET_TIME:
    if (copy_from_user(&rtctm, (struct rtc_time*)arg, sizeof(rtctm))) {
      return -EFAULT;
    }
    return set_rtc_time(&rtctm);
  }
  
  return -EINVAL;
}

static struct file_operations rtc_fops = {
  ioctl: rtc_ioctl,
};

static struct miscdevice rtc_dev = {
  minor: RTC_MINOR,
  name:	 "rtc",
  fops:	 &rtc_fops,
};

static struct i2c_client *eeprom_client = NULL;

int eeprom_ioctl(struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
  if (eeprom_client) {
    return eeprom_client->driver->command(eeprom_client, cmd, (void*)arg);
  }
  
  return -EINVAL;
}

static struct file_operations eeprom_fops = {
  ioctl: eeprom_ioctl,
};

static void armadillo9_setscl(void *data, int state)
{
  if (state) {
    outl(inl(GPIO_PBDR) | 0x20, GPIO_PBDR);
  }
  else {
    outl(inl(GPIO_PBDR) & ~0x20, GPIO_PBDR);
  }
}

static void armadillo9_setsda(void *data, int state)
{
  if (state) {
    outl(inl(GPIO_PBDDR) & ~0x10, GPIO_PBDDR);
  }
  else {
    outl(inl(GPIO_PBDR) & ~0x10, GPIO_PBDR);
    outl(inl(GPIO_PBDDR) | 0x10, GPIO_PBDDR);
  }
}

static int armadillo9_getscl(void *data)
{
  return (inl(GPIO_PBDR) >> 5) & 1;
}

static int armadillo9_getsda(void *data)
{
  return (inl(GPIO_PBDR) >> 4) & 1;
}

static struct i2c_algo_bit_data armadillo9_data = {
  setsda:  armadillo9_setsda,
  setscl:  armadillo9_setscl,
  getsda:  armadillo9_getsda,
  getscl:  armadillo9_getscl,
  udelay:  20,
  mdelay:  20,
  timeout: 100
};

static int armadillo9_client_reg(struct i2c_client *client)
{
  unsigned char ctrl;
  struct rtc_time now;
  
  switch (client->id) {
  case I2C_DRIVERID_S3531A:
    if (client->addr == I2C_S3531A_ADDRESS && !rtc_client) {
      rtc_client = client;
      if (rtc_sync()) {
	printk (KERN_WARNING CHRDEV ": RTC stopped.\n");
	rtc_client = NULL;
	return -1;
      }
      if (get_rtc_time(&now)) {
	printk (KERN_WARNING CHRDEV ": RTC read failed.\n");
	rtc_client = NULL;
	return -1;
      }
      xtime.tv_usec = 0;
      xtime.tv_sec = mktime(1900 + now.tm_year,
			    now.tm_mon + 1, now.tm_mday,
			    now.tm_hour, now.tm_min, now.tm_sec);
    }
    break;
  case I2C_DRIVERID_AT24Cxx:
    if (client->addr == I2C_EEPROM_ADDRESS && !eeprom_client) {
      eeprom_client = client;
    }
    break;
  }
  
  return 0;
}

static int armadillo9_client_unreg(struct i2c_client *client)
{
  if (client == rtc_client && client->addr == I2C_S3531A_ADDRESS) {
    rtc_client = NULL;
  }
  if (client == eeprom_client && client->addr == I2C_EEPROM_ADDRESS) {
    eeprom_client = NULL;
  }
  
  return 0;
}

static struct i2c_adapter armadillo9_ops = {
  name:		     "ARMADILLO9",
  id:		     I2C_HW_B_ARMADILLO9,
  algo_data:	     &armadillo9_data,
  client_register:   armadillo9_client_reg,
  client_unregister: armadillo9_client_unreg,
};

int i2c_armadillo9_init(void)
{
  char buf;
  int  ret;
  
  buf = 0;
  outl(inl(GPIO_PBDR) | 0x20, GPIO_PBDR);
  outl((inl(GPIO_PBDDR) & ~0x10) | 0x20, GPIO_PBDDR);
  
  ret = i2c_bit_add_bus(&armadillo9_ops);
  if (ret < 0) {
    printk (KERN_WARNING CHRDEV ": Unable to add %s bus.\n",
	    armadillo9_ops.name);
    return ret;
  }
  
  if (misc_register(&rtc_dev)) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    rtc_dev.name);
    return ret;
  }
  
  if (register_chrdev (EEPROM_MAJOR, "eeprom", &eeprom_fops)) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    "EEPROM");
    return ret;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

void i2c_armadillo9_cleanup(void)
{
  misc_deregister(&rtc_dev);
  unregister_chrdev (EEPROM_MAJOR, "eeprom");
  i2c_bit_del_bus(&armadillo9_ops);
}

#ifdef MODULE
int init_module(void)
{
  return i2c_armadillo9_init();
}

void cleanup_module(void)
{
  i2c_armadillo9_cleanup();
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
#endif
